unit MasterUnit;

interface

uses
	ProcessUnit, ParsingUnit,
	LinkImportUnit,
	Graphics, SourceEditUnit;

const
	IDELocation = 'Bin\';
	GCCLocation = 'Bin\';
	AsLocation = 'Bin\';
	A68kLocation = 'Bin\';
	ObjCopyLocation = 'Bin\';
	LinkLocation = 'Bin\';
	LdLocation = 'Bin\';
	ArLocation = 'Bin\';
	PackLocation = 'Bin\';
	ASMIncludeLocation = 'Include\ASM\';
	GASIncludeLocation = 'Include\S\';
	CIncludeLocation = 'Include\C\';
	QuillIncludeLocation = 'Include\Quill\';
	DocLocation = 'Doc\';
	ProjectsLocation = 'Projects\';
	StdLibLocation = 'Lib\';
	TIPatchLocation = 'Lib\';
	PStarterLocation = 'Lib\';

	DefaultGCCSwitches = '-B .\ -I- -I ..\Include\C\';
	DefaultAsSwitches = '-mc68000 -I ..\Include\S\';
	DefaultA68kSwitches = '-i..\Include\ASM';
	DefaultObjCopySwitches = '-O coff-m68k';
	DefaultLdSwitches = '-r -e 0';
	DefaultLinkSwitches = '';
	DefaultArSwitches = '-rc';
	SpecialQuillGCCSwitches = '-I ..\Include\Quill\';

	ProjectFileExt = '.tpr';

type
	TSpecialSupportOption = (ssA68k, ssQuill, ssPack, ssAr);
	TSpecialSupport = set of TSpecialSupportOption;
	TProjectTarget = (ptExecutable, ptArchive);
	TTransferTarget = (ttNone, ttVTI, ttCalc);
	TCurVTIType = (cvNone, cvTI89, cvTI92Plus);

var
	MainConsole: TProcessConsole;
	StopOnErrors,
	DeleteAssemblyFiles,
	DeleteObjectFiles,
	AssumeUndefined,
	AutoBlocks,
	AutoNews,
	Pack,
	DebugInfo,
	StdLib: Boolean;
	ProjectTarget: TProjectTarget;
	GCCSwitches,
	AsSwitches,
	AsmSwitches,
	LdSwitches,
	LinkSwitches,
	PackVar,
	CommandLine,
	PostBuildProcessFile: string;
	ProxyName: string;
	ProxyPort: Integer;
	TIGCCFolder: string;
	Temp: string;
	OperationCancelled,
	OperationSuccessful: Boolean;
	ProjectFileName: string;
	PatchLib,
	PatchMainLib: TPatch;
	SyntaxC,
	SyntaxAsmGNU,
	SyntaxAsm,
	SyntaxQuill: TSyntaxColoring;
	TabSizeC,
	TabSizeAsm: Integer;
	EditorColor: TColor;
	EditorFont: TFont;
	EditorOnFly,
	EditorDragDrop,
	EditorRemoveTrSp: Boolean;
	LastNewsDate: Integer;
	TransferTarget: TTransferTarget;
	VTIPath: string;
	LinkPort: TLinkPort;
	LinkCable: TLinkType;
	SpecialSupport: TSpecialSupport;
	Compiling,
	CompFinishAndStop,
	CompStopNow: Boolean;
	CompStartTime,
	CompFileStartTime,
	CompLastTime: TDateTime;
	CompStartFile: procedure of object;
	CompStop: procedure of object;
	CompSetMessage: procedure(const Msg: string) of object;
	CompUpdate: procedure of object;
	CompUpdateProgramOutput: procedure of object;

{$IFDEF CanSplit}
var
	SplitFiles: Boolean;
{$ENDIF}

function StartsWith(const SubStr: string; var Str: string; CaseSensitive: Boolean = False; Remove: Boolean = False): Boolean;

procedure WaitForMainConsole(const ProcessDesc: string);

function Runnable: Boolean;

implementation

uses
	SysUtils, Forms;

procedure WaitForMainConsole(const ProcessDesc: string);
begin
	CompUpdate;
	while MainConsole.Running do begin
		CompUpdate;
		Application.ProcessMessages;
		if CompFinishAndStop then begin
			OperationCancelled := True;
			CompSetMessage ('Finishing ' + ProcessDesc + ' Process');
		end;
		if CompStopNow then begin
			OperationCancelled := True;
			MainConsole.KillProcess;
			Break;
		end;
	end;
	CompUpdate;
end;

function Runnable: Boolean;
begin
	Result := (ProjectTarget <> ptArchive) and (TransferTarget <> ttNone);
end;

function StartsWith(const SubStr: string; var Str: string; CaseSensitive: Boolean; Remove: Boolean): Boolean;
begin
	if CaseSensitive then
		Result := Copy (Str, 1, Length (SubStr)) = SubStr
	else
		Result := UpperCase (Copy (Str, 1, Length (SubStr))) = UpperCase (SubStr);
	if Remove and Result then
		Delete (Str, 1, Length (SubStr));
end;

initialization
	MainConsole := TProcessConsole.Create;
	PatchLib := TPatch.Create;
	PatchLib.Used := True;
	PatchMainLib := TPatch.Create;
	PatchMainLib.Used := True;
	SyntaxC := TSyntaxColoring.Create (nil);
	SyntaxAsm := TSyntaxColoring.Create (nil);
	SyntaxAsmGNU := TSyntaxColoring.Create (nil);
	SyntaxQuill := TSyntaxColoring.Create (nil);
	EditorFont := TFont.Create;
	with SyntaxC do begin
		Enabled := True;
		SymbolColor := clOlive;
		SymbolStyle := [fsBold];
		SymbolCustomStyle := True;
		NumberColor := clMaroon;
		NumberStyle := [];
		NumberCustomStyle := False;
		with WordLists.Add do begin
			Caption := 'C Keywords';
			Words.CommaText :=
				'__alignof__,__asm__,__attribute__,__complex__,__const__,__extension__,__imag__,__inline__,__label__,__real__,__typeof__,' +
				'asm,auto,break,case,char,const,continue,default,do,double,else,enum,extern,float,for,goto,if,inline,int,long,register,return,short,signed,sizeof,static,struct,switch,typedef,typeof,union,unsigned,void,volatile,while';
			CaseSensitive := True;
			CustomColor := True;
			Color := clBlue;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with CustomStyles.Add do begin
			Caption := 'Comment Area';
			CustomColor := True;
			Color := clGreen;
			CustomStyle := True;
			Style := [fsItalic];
			BeginText := '/*';
			EndText := '*/';
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'Comment Line';
			CustomColor := True;
			Color := clGreen;
			CustomStyle := True;
			Style := [fsItalic];
			BeginText := '//';
			EndText := #13;
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'String';
			CustomColor := True;
			Color := clMaroon;
			CustomStyle := False;
			Style := [];
			BeginText := '"';
			EndText := '"';
			IgnoreChar := '\';
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'Character';
			CustomColor := True;
			Color := clMaroon;
			CustomStyle := False;
			Style := [];
			BeginText := '''';
			EndText := '''';
			IgnoreChar := '\';
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'Compiler Directive';
			CustomColor := True;
			Color := clTeal;
			CustomStyle := True;
			Style := [fsBold];
			BeginText := '#';
			EndText := #13;
			IgnoreChar := '\';
			Switchable := True;
		end;
		ParenthesisColors.CommaText := '$800080,$C08000,$8080FF,$008000';
		ParenthesisStyle := [fsBold];
		ParenthesisCustomStyle := True;
	end;
	with SyntaxAsmGNU do begin
		NumberColor := SyntaxC.NumberColor;
		NumberStyle := SyntaxC.NumberStyle;
		NumberCustomStyle := SyntaxC.NumberCustomStyle;
		SymbolColor := SyntaxC.SymbolColor;
		SymbolStyle := SyntaxC.SymbolStyle;
		SymbolCustomStyle := SyntaxC.SymbolCustomStyle;
		ParenthesisColors.Assign (SyntaxC.ParenthesisColors);
		ParenthesisStyle := SyntaxC.ParenthesisStyle;
		ParenthesisCustomStyle := SyntaxC.ParenthesisCustomStyle;
		with CustomStyles.Add do begin
			Caption := 'Comment Area';
			CustomColor := True;
			Color := clGreen;
			CustomStyle := True;
			Style := [fsItalic];
			BeginText := '/*';
			EndText := '*/';
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'Comment Line (|)';
			CustomColor := True;
			Color := clGreen;
			CustomStyle := True;
			Style := [fsItalic];
			BeginText := '|';
			EndText := #13;
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'Comment Line (#)';
			CustomColor := True;
			Color := clGreen;
			CustomStyle := True;
			Style := [fsItalic];
			BeginText := '#';
			EndText := #13;
			Switchable := False;
			LineStartOnly := True;
		end;
		with CustomStyles.Add do begin
			Caption := 'String';
			CustomColor := True;
			Color := clMaroon;
			CustomStyle := False;
			Style := [];
			BeginText := '"';
			EndText := '"';
			IgnoreChar := '\';
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'Character';
			CustomColor := True;
			Color := clMaroon;
			CustomStyle := False;
			Style := [];
			BeginText := '''';
			EndText := '''';
			IgnoreChar := '\';
			Switchable := False;
		end;
		with WordLists.Add do begin
			Caption := 'Data Movement';
			Words.CommaText := 'EXG,LEA,LINK,MOV,MOVE,MOVEA,MOVEM,MOVEP,MOVEQ,MOVM,MOVP,MOVQ,PEA,UNLK';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Integer Arithmetic';
			Words.CommaText := 'ADD,ADDA,ADDI,ADDQ,ADDX,CLR,CMP,CMPA,CMPI,CMPM,DIVS,DIVU,EXT,MULS,MULU,NEG,NEGX,SUB,SUBA,SUBI,SUBQ,SUBX';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Logical Instructions';
			Words.CommaText := 'AND,ANDI,EOR,EORI,NOT,OR,ORI';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Shift/Rotation Instructions';
			Words.CommaText := 'ASL,ASR,LSL,LSR,ROL,ROR,ROXL,ROXR,SWAP';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Bit Manipulation';
			Words.CommaText := 'BCHG,BCLR,BSET,BTST';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Program Control';
			Words.CommaText := 'BCC,BCS,BEQ,BGE,BGT,BHI,BLE,BLS,BLT,BMI,BNE,BPL,BRA,BSR,BVC,BVS,JBCC,JBCS,JBEQ,JBGE,JBGT,JBHI,JBLE,JBLS,JBLT,JBMI,JBNE,JBPL,JBRA,JBSR,JBVC,JBVS,JSR,JRA,JMP,'
				+ 'NOP,RTR,RTS,SCC,SCS,SEQ,SF,SGE,SGT,SHI,SLE,SLS,SLT,SMI,SNE,SPL,ST,SVC,SVS,TST,'
				+ 'JHI,JLS,JCC,JCS,JNE,JEQ,JVC,JVS,JPL,JMI,JGE,JLT,JGT,JLE,DBHI,DBLS,DBCC,DBCS,DBNE,DBEQ,DBVC,DBVS,DBPL,DBMI,DBGE,DBLT,DBGT,DBLE,DBF,DBRA,DBT,'
				+ 'FJNE,FJEQ,FJGE,FJLT,FJGT,FJLE,FJF,FJT,FJGL,FJGLE,FJNGE,FJNGL,FJNGLE,FJNGT,FJNLE,FJNLT,FJOGE,FJOGL,FJOGT,FJOLE,FJOLT,FJOR,FJSEQ,FJSF,FJSNE,FJST,FJUEQ,FJUGE,FJUGT,FJULE,FJULT,FJUN';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'System Control';
			Words.CommaText := 'ILLEGAL,RTE,TRAP';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Extensions';
			Words.CommaText := 'B,L,S,W';
			Color := $408000;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Assembler Directives';
			Words.CommaText := 'abort,align,ascii,asciz,balign,balignw,balignl,byte,comm,data,'
				+ 'def,dim,double,eject,else,end,elseif,endef,endfunc,endif,endm,endr,equ,equiv,'
				+ 'err,even,exitm,extern,fail,file,fill,float,func,global,hword,ident,if,include,'
				+ 'incbin,int,irp,irpc,lcomm,lflags,line,ln,list,long,macro,mri,nolist,octa,org,'
				+ 'p2align,p2alignw,p2alignl,print,psize,purgem,quad,rept,sbttl,scl,section,set,'
				+ 'short,single,size,sleb128,skip,space,stabd,stabn,stabs,string,struct,tag,text,'
				+ 'title,type,uleb128,val,vtable_entry,word';
			Color := clBlue;
			CustomColor := True;
			Style := [fsBold];
			CustomStyle := True;
			CaseSensitive := True;
		end;
		with WordLists.Add do begin
			Caption := 'Registers';
			Words.CommaText := 'a0,a1,a2,a3,a4,a5,a6,a7,d0,d1,d2,d3,d4,d5,d6,d7,fp,pc,sp,sr';
			Color := clRed;
			CustomColor := True;
			Style := [fsUnderline];
			CustomStyle := True;
			CaseSensitive := False;
		end;
	end;
	with SyntaxAsm do begin
		NumberColor := SyntaxC.NumberColor;
		NumberStyle := SyntaxC.NumberStyle;
		NumberCustomStyle := SyntaxC.NumberCustomStyle;
		SymbolColor := SyntaxC.SymbolColor;
		SymbolStyle := SyntaxC.SymbolStyle;
		SymbolCustomStyle := SyntaxC.SymbolCustomStyle;
		ParenthesisColors.Assign (SyntaxC.ParenthesisColors);
		ParenthesisStyle := SyntaxC.ParenthesisStyle;
		ParenthesisCustomStyle := SyntaxC.ParenthesisCustomStyle;
		with CustomStyles.Add do begin
			Caption := 'Comment';
			CustomColor := True;
			Color := clGreen;
			CustomStyle := True;
			Style := [fsItalic];
			BeginText := ';';
			EndText := #13;
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'String (double-quoted)';
			CustomColor := True;
			Color := clMaroon;
			CustomStyle := False;
			Style := [];
			BeginText := '"';
			EndText := '"';
			IgnoreChar := '';
			Switchable := False;
		end;
		with CustomStyles.Add do begin
			Caption := 'String (single-quoted)';
			CustomColor := True;
			Color := clMaroon;
			CustomStyle := False;
			Style := [];
			BeginText := '''';
			EndText := '''';
			IgnoreChar := '';
			Switchable := False;
		end;
		with WordLists.Add do begin
			Caption := 'Data Movement';
			Words.CommaText := 'EXG,LEA,LINK,MOV,MOVE,MOVEA,MOVEM,MOVEP,MOVEQ,PEA,UNLK';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Integer Arithmetic';
			Words.CommaText := 'ADD,ADDA,ADDI,ADDQ,ADDX,CLR,CMP,CMPA,CMPI,CMPM,DIVS,DIVU,EXT,MULS,MULU,NEG,NEGX,SUB,SUBA,SUBI,SUBQ,SUBX';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Logical Instructions';
			Words.CommaText := 'AND,ANDI,EOR,EORI,NOT,OR,ORI';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Shift/Rotation Instructions';
			Words.CommaText := 'ASL,ASR,LSL,LSR,ROL,ROLX,ROR,RORX,ROXL,ROXR,SWAP';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Bit Manipulation';
			Words.CommaText := 'BCHG,BCLR,BSET,BTST';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Program Control';
			Words.CommaText := 'BCC,BCS,BEQ,BGE,BGT,BHI,BHS,BLE,BLO,BLS,BLT,BMI,BNE,BPL,BRA,BSR,BVC,BVS,DBCC,DBCS,DBEQ,DBF,DBGE,DBGT,DBHI,DBHS,DBLE,DBLO,DBLS,DBLT,DBMI,DBNE,DBPL,DBRA,DBT,DBVC,DBVS,JMP,JSR,NOP,RTR,RTS,'
				+ 'SCC,SCS,SEQ,SF,SGE,SGT,SHI,SHS,SLE,SLO,SLS,SLT,SMI,SNE,SPL,ST,SVC,SVS,TST';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'System Control';
			Words.CommaText := 'ILLEGAL,RTE,TRAP';
			Color := clBlue;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Extensions';
			Words.CommaText := 'B,L,S,W';
			Color := $408000;
			CustomColor := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Assembler Directives';
			Words.CommaText := 'BSS,CNOP,CSEG,DSEG,DC,DCB,DS,END,ENDC,ENDIF,ENDM,EQU,EQUR,EVEN,FAR,IDNT,IFC,IFD,IFEQ,IFGE,IFGT,IFLE,IFLT,IFNC,IFND,IFNE,INCBIN,INCLUDE,LIST,MACRO,NEAR,NOLIST,ORG,PAGE,PUBLIC,REG,RORG,SECTION,SET,SPC,TITLE,TTL,XDEF,XREF';
			Color := clBlue;
			CustomColor := True;
			Style := [fsBold];
			CustomStyle := True;
			CaseSensitive := False;
		end;
		with WordLists.Add do begin
			Caption := 'Registers';
			Words.CommaText := 'a0,a1,a2,a3,a4,a5,a6,a7,d0,d1,d2,d3,d4,d5,d6,d7,fp,pc,sp,sr';
			Color := clRed;
			CustomColor := True;
			Style := [fsUnderline];
			CustomStyle := True;
			CaseSensitive := False;
		end;
	end;
	SyntaxQuill.Assign (SyntaxC);
	with SyntaxQuill do begin
		with WordLists.Add do begin
			Caption := 'Sections';
			Words.CommaText := '$$ACTIONS,$$CONNECTIONS,$$END,$$END_TEST,$$EVENTS,$$EXTERN,$$LOCATIONS,$$MESSAGES,$$OBJECTS,$$OLDSTYLE_SYSTEM_MESSAGES,$$PICTURES,$$PICTURES_TEST,$$SYSTEM_MESSAGES,$$TITLE,$$VOCABULARY';
			CaseSensitive := True;
			CustomColor := True;
			Color := $0000FF;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Section-specific Keywords';
			Words.CommaText := 'ACTION,BITMAP,CBLOCK,CONN,DEFINE,DRAWING,END_BITMAP,END_CBLOCK,END_DRAWING,END_PACKED_BITMAP,EVENT,FROM,LOC,MSG,OBJ,PACKED_BITMAP,WORD';
			CaseSensitive := True;
			CustomColor := True;
			Color := $808040;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Additional Keywords';
			Words.CommaText := 'CONTINUE,ELSE';
			CaseSensitive := True;
			CustomColor := True;
			Color := $808040;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Non-functional Keywords';
			Words.CommaText := 'BEEP,BLOCK,BORDER,BRIGHT,FLASH,INK,PAPER';
			CaseSensitive := True;
			CustomColor := True;
			Color := $C0C0C0;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Predefined Aliases';
			Words.CommaText := '$ALSOSEE,$ARG,$CARRIED,$CENTER,$CNT1,$CNT2,$CNT3,$CNT4,$CONT,$CURLOC,$DARK,$DARKCNT,$DESC,$DESC_CNT,$DESC_DARKCNT,$DESC_NOLIGHTCNT,$DONE,$ENDGAME,$EXIT,$FAIL,$FONT,$FULLSCR,$GCONTROL,$LSOURCE,'
				+ '$MAXCAR,$NOLIGHTCNT,$NOUN,$NOWHERE,$NULL,$NUMCAR,$PROMPT,$RESTART,$SCORE,$SPECIAL,$SUBROUTINE,$TURNHI,$TURNLO,$VERB,$WORN';
			CaseSensitive := True;
			CustomColor := True;
			Color := $800080;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Conditions';
			Words.CommaText := 'ABSENT,AT,ATGT,ATLT,CARRIED,CHANCE,CREATED,EQ,EQWORD,EXTWORD,GT,HERE,ISAT,ISDESC,ISNOTAT,ISNOTNULL,ISNULL,LT,NEQWORD,NOTAT,NOTCARR,NOTCREATED,NOTEQ,NOTHERE,NOTSAME,NOTWORN,NOTZERO,PRESENT,SAME,TRYMOVE,WORN,ZERO';
			CaseSensitive := True;
			CustomColor := True;
			Color := $FF0000;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Actions';
			Words.CommaText := 'ADD,ALSOSEE,ANYKEY,AUTOD,AUTOG,AUTOR,AUTOW,BIGFONT,CANCEL,CLEAR,CLS,COPYFF,COPYFO,COPYOF,COPYOO,CREATE,DECCAR,DESC,DESTROY,DONE,DROP,DROPALL,END,ENDDESC,EXIT,EXTERN,GET,GETWORD,GOTO,INCCAR,INVEN,LET,'
				+ 'LISTAT,LISTOBJ,LOAD,MAXCAR,MES,MESFLAG,MESSAGE,MINUS,NEWLINE,NOTDONE,OK,PAUSE,PICNORM,PICOFF,PICON,PLACE,PLUS,PRINT,PROMPT,PUTO,QUIT,QVERSION,RAMLOAD,RAMSAVE,RANDOM,REDRAW,REMOVE,RESTART,SAVE,SCORE,'
				+ 'SET,SETNOUN,SETVERB,SHOWLOC,SMLFONT,SUB,SWAP,SYSMESS,TURNS,WEAR,WHATO,WHEREO,ZAPSCR';
			CaseSensitive := True;
			CustomColor := True;
			Color := $A00000;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Drawing Primitives';
			Words.CommaText := 'AMOVE,CALL,ELLIPSE,FILL,INV_ELLIPSE,INV_LINE,INV_PLOT,INV_RPLOT,LINE,MOVE,PLOT,RPLOT,SHADE,XOR_ELLIPSE,XOR_LINE,XOR_PLOT,XOR_RPLOT';
			CaseSensitive := True;
			CustomColor := True;
			Color := $FF8000;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Drawing Directions';
			Words.CommaText := 'DOWN,DOWN_LEFT,DOWN_RIGHT,LEFT,LEFT_DOWN,LEFT_UP,RIGHT,RIGHT_DOWN,RIGHT_UP,UP,UP_LEFT,UP_RIGHT';
			CaseSensitive := True;
			CustomColor := True;
			Color := $008000;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'Shading Patterns';
			Words.CommaText := '$BKSLASHFILL,$BRICKFILL,$CHAINFILL,$CIRCLEFILL,$CLOSEDASHFILL,$CLOSEDOTFILL,$CLOSEWAVEFILL,$CROSSFILL,$DASHFILL,$DOTFILL,$HATCHFILL,$INTERLEAVEFILL,$LIGHTDOTFILL,$LIGHTLINEFILL,$LINEFILL,$SLASHFILL,'
				+ '$SOLIDFILL,$SQDOTFILL,$SQUAREFILL,$THICKBKSLASHFILL,$THICKHATCHFILL,$THICKLINEFILL,$THICKSLASHFILL,$VDASHFILL,$VINTERLEAVEFILL,$VLDOTFILL,$VLIGHTLINEFILL,$VLINEFILL,$VTHICKLINEFILL,$WAVEFILL,$WIDEDOTFILL,$XMARKFILL,$ZIGZAGFILL';
			CaseSensitive := True;
			CustomColor := True;
			Color := $800080;
			CustomStyle := True;
			Style := [fsBold];
		end;
		with WordLists.Add do begin
			Caption := 'External Symbols';
			Words.CommaText := '$ACTIONS$,$ARG$,$BMPUT$,$BPCKPUT$,$BUFFER$,$CONNECTIONS$,$EVENTS$,$EXTERN$,$FLAGS$,$FLAGS_BACKUP$,$GDF$,$GETLINE$,$LOCATIONS$,$LQL$,$MAXCAR$,$MESSAGES$,$NFLAG$,$NLOC$,$NMSG$,$NOBJ$,$NSYSMSG$,$NWORD$,'
				+ '$OBJECTS$,$PDRAW$,$PICTURE$,$PRINT$,$RAM_SAVED$,$SCALEX$,$SCALEY$,$SSCR$,$SYSTEM_MESSAGES$,$WORDS$';
			CaseSensitive := True;
			CustomColor := True;
			Color := $408000;
			CustomStyle := True;
			Style := [fsBold];
		end;
	end;
	EditorFont.Name := 'Courier New';
	EditorFont.Size := 10;
	EditorColor := clWindow;
	DecimalSeparator := '.';
finalization
	EditorFont.Free;
	SyntaxQuill.Free;
	SyntaxAsm.Free;
	SyntaxAsmGNU.Free;
	SyntaxC.Free;
	PatchMainLib.Free;
	PatchLib.Free;
	MainConsole.Free;
end.
