unit ParsingUnit;

interface

uses
	ObjectFileUnit,
	SysUtils, Classes;

const
	LineIndicator = '#L';
	OffsetIndicator = '#$';
	PatchToken = '$$PATCH';
	ErrorToken = '$$ERROR';
	WarningToken = '$$WARNING';
	ALineIdentifier = '_A_LINE';
	FLineIdentifier = '_F_LINE';
	SParseError = 'Line %d: Parse error before ''%s''';

type
	EFileParseError = class(Exception);

	PBoolean = ^Boolean;

	TCalcDest = (cdTI89, cdTI92Plus);
	TCalcDests = set of TCalcDest;

	TPatch = class(TStringList)
	private
		FGlobalPart: TStringList;
		FUsed: Boolean;
	public
		constructor Create;
		destructor Destroy; override;
		procedure Clear; override;
		procedure LoadFromFile(const FileName: string); override;
		procedure UseKeyWord(const KeyWord: string); virtual;
		procedure DeselectAll; virtual;
		function InsertInto(Dest: TStringList; var LineIndex: Integer; NeedFLine: PBoolean = nil): Boolean; virtual;
	published
		property GlobalPart: TStringList read FGlobalPart;
		property Used: Boolean read FUsed write FUsed;
	end;

	TLineChangeEvent = procedure(StartLine, Change: Integer) of object;
	TLineContentsEvent = function(StartLine, EndLine: Integer): string of object;

	TTFRes2 = array [0..1] of Byte;
	TTFRes3 = array [0..2] of Byte;
	TTFRes6 = array [0..5] of Byte;

	TTransferFileHeader = record
		Signature: array [0..7] of Char; // "**TI92P*" or "**TI89**"
		Reserved1: TTFRes2;              // 01 00
		Folder: array [0..7] of Char;    // folder name
		Desc: array [0..39] of Char;     // not used
		Reserved2: TTFRes6;              // 01 00 52 00 00 00
		VarName: array [0..7] of Char;   // variable name
		LinkType: TPCInt1;               // variable link type (0C = string, 1C = other, 21 = program, ...)
		Reserved3: TTFRes3;              // 00 00 00
		FileSize: TPCInt4;               // file size from Signature to CheckSum
		Reserved4: TTFRes6;              // A5 5A 00 00 00 00
	end;
	PTransferFileHeader = ^TTransferFileHeader;

	TCalcVarHeader = record
		DataSize: TCalcInt2;             // data size (including Tag)
	end;
	PCalcVarHeader = ^TCalcVarHeader;

	TCalcVarFooter = record
		Tag: TCalcInt1;                  // variable tag
	end;
	PCalcVarFooter = ^TCalcVarFooter;

	TTransferFileFooter = record
		CheckSum: TPCInt2;               // checksum from DataSize to Tag
	end;
	PTransferFileFooter = ^TTransferFileFooter;

	PBinData = PChar;

const
	MaxCalcAllocBlock = $FFF0;

procedure ParseSFile(Contents: TStringList; PatchLib, PatchMainLib: TPatch; out MainPatchIncluded: Boolean);
procedure ParseDebugSFile(Contents: TStringList; OnLineChange: TLineChangeEvent; LineContents: TStringList; OnGetLineContents: TLineContentsEvent);
procedure ParseCOFFDebugSFile(Contents: TStringList; const Header: TCOFFSectionHeader; const SourceFileName: string; const Symbols: TCOFFSymbols; const Strings: TStream; const Lines: TCOFFLines; out FileLines: TCOFFFileLines);
procedure ParsePStarter(const InputFile: string; const OutputFile: string; const PackVar: string);

procedure ParseDebugSFileDisk(Contents: TStringList; const SourceFileName: string);
procedure ParseDebugFiles(DebugFileNames: TStringList; const MergedObjectFileName: string);

function GetCalcVarSize(ContentLength: LongWord; const Extension: string = ''): LongWord;
function GetTransferFileSize(ContentLength: LongWord; const Extension: string = ''): LongWord;
procedure ProduceTransferFile(Output: Pointer; Contents: Pointer; ContentLength: LongWord; CalcDest: TCalcDest; const DestFolder, DestVarName: string; VarLinkType, VarTag: Byte; const Extension: string = '');

var
	ErrorMessageProc: procedure(const Msg: string) = nil;

implementation

uses
	Windows;

{ General parsing procedures }

procedure ParseSFile(Contents: TStringList; PatchLib, PatchMainLib: TPatch; out MainPatchIncluded: Boolean);
var
	I,
	P1,
	P2: Integer;
	PrevPatchLine,
	PatchLine,
	MainPatchLine: Integer;
	PatchFound,
	MainPatchFound,
	RegRelative,
	NeedFLine: Boolean;
	L,
	W: string;
begin
	PatchLine := 0;
	PatchFound := False;
	MainPatchLine := 0;
	MainPatchFound := False;
	RegRelative := False;
	with Contents do
		for I := 0 to Count - 1 do begin
			L := Strings [I];
			if (L = '__relation:') or (L = #9'.macro __init_reg_relative') then
				RegRelative := True;
			if (L = 'patch_start:') or (L = '__relation:') or ((not PatchFound) and (L = 'tigcc_compiled.:')) then begin
				PatchLine := I + 1;
				PatchFound := True;
			end;
			if (L = '_main:') or (L = '__main:') then begin
				MainPatchLine := I + 1;
				MainPatchFound := True;
			end;
			if RegRelative then begin
				P1 := Pos ('-__relation', L);
				if P1 > 0 then begin
					P2 := Pos ('_CALL_', L);
					if (P2 > 0) and (P2 < P1) and (P1 - P2 <= Length ('_CALL_') + 4) then begin
						System.Delete (L, P1, Length ('-__relation') + 1);
						while (Length (L) > 0) and (L [P1] <> ')') do
							System.Delete (L, P1, 1);
						System.Delete (L, P1, 1);
						Strings [I] := L;
					end;
					P2 := Pos ('_ER_CODE_', L);
					if (P2 > 0) and (P2 < P1) and (P1 - P2 <= Length ('_ER_CODE_') + 5) then begin
						System.Delete (L, P1, Length ('-__relation') + 1);
						while (Length (L) > 0) and (L [P1] <> ')') do
							System.Delete (L, P1, 1);
						System.Delete (L, P1, 1);
						Strings [I] := L;
					end;
				end;
			end;
			while Length (L) > 0 do begin
				W := '';
				while (Length (L) > 0) and (L [1] in ['A'..'Z', 'a'..'z', '0'..'9', '_']) do begin
					System.Insert (L [1], W, Length (W) + 1);
					System.Delete (L, 1, 1);
				end;
				if Length (W) > 0 then begin
					PatchLib.UseKeyWord (W);
					PatchMainLib.UseKeyWord (W);
				end;
				System.Delete (L, 1, 1);
			end;
		end;
	PrevPatchLine := PatchLine;
	NeedFLine := False;
	PatchLib.InsertInto (Contents, PatchLine, @NeedFLine);
	if PrevPatchLine <= MainPatchLine then
		Inc (MainPatchLine, PatchLine - PrevPatchLine);
	if MainPatchFound then
		MainPatchIncluded := PatchMainLib.InsertInto (Contents, MainPatchLine, @NeedFLine);
	with Contents do begin
		if NeedFLine then
			for I := 0 to Count - 1 do begin
				L := Strings [I];
				if Copy (L, 1, Length (#9'jbsr _ROM_CALL_')) = #9'jbsr _ROM_CALL_' then begin
					System.Delete (L, 1, Length (#9'jbsr _ROM_CALL_'));
					if Length (L) <= 3 then
						Strings [I] := #9'.word _F_LINE+0x' + L;
				end else if Copy (L, 1, Length (#9'jsr _ROM_CALL_')) = #9'jsr _ROM_CALL_' then begin
					System.Delete (L, 1, Length (#9'jsr _ROM_CALL_'));
					if Length (L) <= 3 then
						Strings [I] := #9'.word _F_LINE+0x' + L;
				end;
			end;
		for I := 0 to Count - 1 do begin
			L := Strings [I];
			if Copy (L, 1, Length (#9'jra _ER_CODE_')) = #9'jra _ER_CODE_' then begin
				System.Delete (L, 1, Length (#9'jra _ER_CODE_'));
				if Length (L) <= 4 then
					Strings [I] := #9'.word _A_LINE+' + L;
			end else if Copy (L, 1, Length (#9'jmp _ER_CODE_')) = #9'jmp _ER_CODE_' then begin
				System.Delete (L, 1, Length (#9'jmp _ER_CODE_'));
				if Length (L) <= 4 then
					Strings [I] := #9'.word _A_LINE+' + L;
			end;
		end;
	end;
end;

procedure ParseDebugSFile(Contents: TStringList; OnLineChange: TLineChangeEvent; LineContents: TStringList; OnGetLineContents: TLineContentsEvent);
function GetLineContents(StartLine, EndLine: Integer): string;
var
	I: Integer;
begin
	if Assigned (OnGetLineContents) then
		Result := OnGetLineContents (StartLine, EndLine)
	else if Assigned (LineContents) then begin
		Result := '';
		for I := StartLine - 1 to EndLine - 1 do
			Insert (LineContents.Strings [I] + #13#10, Result, Length (Result) + 1);
		Delete (Result, Length (Result) - 1, 2);
	end else
		Result := '';
end;
var
	I,
	P,
	LastLine,
	RecontinueLine,
	LastDebugLine,
	CurDebugLine: Integer;
	S: string;
begin
	LastLine := -1;
	LastDebugLine := -1;
	RecontinueLine := -1;
	I := 0;
	with Contents do
		while I <= Count do begin
			if (I < Count) and (Copy (Strings [I], 1, Length (#9'.def'#9)) = #9'.def'#9) then begin
				Delete (I);
				if Assigned (OnLineChange) then
					OnLineChange (I, -1);
				Dec (I);
			end else if (I >= Count) or (Copy (Strings [I], 1, Length (#9'.ln'#9)) = #9'.ln'#9) then begin
				if I < Count then begin
					S := Strings [I];
					System.Delete (S, 1, Length (#9'.ln'#9));
					try
						CurDebugLine := StrToInt (S);
					except
						CurDebugLine := LastDebugLine;
					end;
					System.Insert (LineIndicator, S, 1);
					System.Insert (#9, S, Length (S) + 1);
					Strings [I] := S;
				end else
					CurDebugLine := LastDebugLine + 1;
				if (LastLine >= 0) and (CurDebugLine >= 0) then begin
					S := Strings [LastLine];
					if CurDebugLine < LastDebugLine then begin
						System.Insert (StringReplace (Trim (GetLineContents (LastDebugLine, LastDebugLine)), #9, '', [rfReplaceAll]) + ' ...', S, Length (S) + 1);
						RecontinueLine := LastDebugLine;
					end else begin
						if (RecontinueLine >= 0) and (CurDebugLine >= RecontinueLine) then begin
							System.Insert (StringReplace (Trim (GetLineContents (LastDebugLine, LastDebugLine)) + ' ...', #9, '', [rfReplaceAll]), S, Length (S) + 1);
							RecontinueLine := -1;
						end else
							System.Insert (StringReplace (Trim (GetLineContents (LastDebugLine, CurDebugLine - 1)), #9, '', [rfReplaceAll]), S, Length (S) + 1);
					end;
					repeat
						P := Pos (#10, S);
						if P <= 0 then
							P := Pos (#13, S);
						if P > 0 then
							S [P] := ' ';
					until P <= 0;
					Strings [LastLine] := S;
				end;
				LastLine := I;
				LastDebugLine := CurDebugLine;
			end;
			Inc (I);
		end;
end;

procedure ParseCOFFDebugSFile(Contents: TStringList; const Header: TCOFFSectionHeader; const SourceFileName: string; const Symbols: TCOFFSymbols; const Strings: TStream; const Lines: TCOFFLines; out FileLines: TCOFFFileLines);
var
	I,
	J,
	Line,
	CurTableIndex: Integer;
	S: string;
begin
	ExtractCOFFFileLines (Header, ExtractFileName (SourceFileName), Symbols, Strings, Lines, FileLines);
	CurTableIndex := Low (FileLines.Lines);
	with Contents do
		for I := 0 to Count - 1 do
			if Copy (Strings [I], 1, Length (LineIndicator)) = LineIndicator then begin
				S := Strings [I];
				System.Delete (S, 1, Length (LineIndicator));
				try
					Line := StrToInt (Copy (S, 1, Pos (#9, S) - 1));
					System.Delete (S, 1, Pos (#9, S) - 1);
					with FileLines do
						for J := CurTableIndex to High (Lines) do
							with Lines [J] do
								if LineNumber = Line then begin
									System.Insert (OffsetIndicator + IntToHex (LineAddress, 5), S, 1);
									Strings [I] := S;
									Inc (CurTableIndex);
									Break;
								end;
				except end;
			end;
end;

procedure ParsePStarter(const InputFile: string; const OutputFile: string; const PackVar: string);
const
	TempProg = 'TEMPPROG';
var
	I: Integer;
	ObjectFile: TMemoryStream;
	Buf: PChar;
	Len: Integer;
	TempBuf: array [0..Length(TempProg)] of Char;
begin
	TempBuf [Length (TempProg)] := #0;
	ObjectFile := TMemoryStream.Create;
	try
		ObjectFile.LoadFromFile (InputFile);
		Buf := ObjectFile.Memory;
		Len := ObjectFile.Size;
		for I := 0 to Len - Length (TempProg) do
			if UpCase (Buf [I]) = TempProg [1] then begin
				Move (Buf [I], TempBuf, Length (TempProg));
				if UpperCase (AnsiString (TempBuf)) = TempProg then begin
					FillChar (Buf [I], Length (TempProg), 0);
					StrPLCopy (PChar (@(Buf [I])), LowerCase (PackVar), Length (TempProg));
					Break;
				end;
			end;
		ObjectFile.SaveToFile (OutputFile);
	finally
		ObjectFile.Free;
	end;
end;

procedure ParseDebugSFileDisk(Contents: TStringList; const SourceFileName: string);
var
	L: TStringList;
begin
	L := TStringList.Create;
	with L do try
		LoadFromFile (SourceFileName);
		ParseDebugSFile (Contents, nil, L, nil);
	finally
		Free;
	end;
end;

procedure ParseDebugFiles(DebugFileNames: TStringList; const MergedObjectFileName: string);
var
	I: Integer;
	L,
	M: TStringList;
	COFFStream: TFileStream;
	COFFHeaders: TCOFFHeaders;
	COFFSymbols: TCOFFSymbols;
	COFFStrings: TMemoryStream;
	COFFLines: TCOFFLines;
	COFFFileLines: TCOFFFileLines;
	CurSection: Integer;
begin
	COFFStrings := TMemoryStream.Create;
	try
		COFFStream := TFileStream.Create (MergedObjectFileName, fmOpenRead or fmShareDenyWrite);
		try
			ReadCOFFHeaders (COFFStream, COFFHeaders);
			ReadCOFFSymbols (COFFStream, COFFHeaders.FileHeader, COFFSymbols);
			ReadCOFFStrings (COFFStream, COFFHeaders.FileHeader, COFFStrings);
			with COFFHeaders do
				for CurSection := Low (SectionHeaders) to High (SectionHeaders) do
					with SectionHeaders [CurSection] do
						if ReadCalcInt (LineCount) > 0 then begin
							ReadCOFFLines (COFFStream, SectionHeaders [CurSection], COFFLines);
							for I := 0 to DebugFileNames.Count - 1 do begin
								L := TStringList.Create;
								with L do try
									LoadFromFile (ChangeFileExt (DebugFileNames.Strings [I], '.s'));
									M := TStringList.Create;
									with M do try
										LoadFromFile (DebugFileNames.Strings [I]);
										ParseDebugSFile (L, nil, M, nil);
									finally
										Free;
									end;
									ParseCOFFDebugSFile (L, SectionHeaders [CurSection], DebugFileNames.Strings [I], COFFSymbols, COFFStrings, COFFLines, COFFFileLines);
									SaveToFile (ChangeFileExt (DebugFileNames.Strings [I], '.s'));
								finally
									Free;
								end;
							end;
							Break;
						end;
		finally
			COFFStream.Free;
		end;
	finally
		COFFStrings.Free;
	end;
end;

{ TPatch }

procedure TPatch.Clear;
var
	I: Integer;
begin
	with GlobalPart do
		for I := 0 to Count - 1 do
			if Assigned (Objects [I]) then
				Objects[I].Free;
	GlobalPart.Clear;
	inherited;
end;

constructor TPatch.Create;
begin
	inherited;
	Sorted := True;
	Duplicates := dupAccept;
	FGlobalPart := TStringList.Create;
end;

procedure TPatch.DeselectAll;
var
	I: Integer;
begin
	Used := False;
	for I := 0 to Count - 1 do
		(Objects [I] as TPatch).Used := False;
end;

destructor TPatch.Destroy;
begin
	Clear;
	FGlobalPart.Free;
	inherited;
end;

function TPatch.InsertInto(Dest: TStringList; var LineIndex: Integer; NeedFLine: PBoolean): Boolean;
var
	I: Integer;
	S: string;
begin
	Result := False;
	with GlobalPart do
		for I := 0 to Count - 1 do
			if Assigned (Objects [I]) then begin
				with Objects [I] as TPatch do
					if Used then begin
						InsertInto (Dest, LineIndex, NeedFLine);
						Result := True;
					end;
			end else begin
				S := Strings [I];
				if Assigned (NeedFLine) and (not NeedFLine^) and (Pos (FLineIdentifier, S) > 0) then
					NeedFLine^ := True;
				if Pos (ErrorToken, UpperCase (S)) > 0 then begin
					if Assigned (ErrorMessageProc) then begin
						System.Delete (S, 1, Pos (ErrorToken, UpperCase (S)) - 1 + Length (ErrorToken));
						S := Trim (S);
						if (Length (S) > 0) and (S [1] = '"') and (S [Length (S)] = '"') then begin
							ErrorMessageProc ('Error: ' + Copy (S, 2, Length (S) - 2));
							Abort;
						end;
					end;
				end else if Pos (WarningToken, UpperCase (S)) > 0 then begin
					if Assigned (ErrorMessageProc) then begin
						System.Delete (S, 1, Pos (WarningToken, UpperCase (S)) - 1 + Length (WarningToken));
						S := Trim (S);
						if (Length (S) > 0) and (S [1] = '"') and (S [Length (S)] = '"') then
							ErrorMessageProc ('Warning: ' + Copy (S, 2, Length (S) - 2));
					end;
				end else begin
					if (LineIndex - 1 >= 0) and (Pos (':', S) > 0) then begin
						if Pos (Trim (Copy (S, 1, Pos (':', S) - 1)), Dest.Strings [LineIndex - 1]) > 0 then begin
							S := LowerCase (Dest.Strings [LineIndex - 1]);
							if (Pos ('jmp', S) > 0)
								or (Pos ('bra', S) > 0)
								or (Pos ('bcc', S) > 0)
								or (Pos ('bcs', S) > 0)
								or (Pos ('bne', S) > 0)
								or (Pos ('beq', S) > 0)
								or (Pos ('bvc', S) > 0)
								or (Pos ('bvs', S) > 0)
								or (Pos ('bpl', S) > 0)
								or (Pos ('bmi', S) > 0)
								or (Pos ('bhi', S) > 0)
								or (Pos ('bls', S) > 0)
								or (Pos ('bge', S) > 0)
								or (Pos ('blt', S) > 0)
								or (Pos ('bgt', S) > 0)
								or (Pos ('ble', S) > 0)
								or (Pos ('jra', S) > 0)
								or (Pos ('jcc', S) > 0)
								or (Pos ('jcs', S) > 0)
								or (Pos ('jne', S) > 0)
								or (Pos ('jeq', S) > 0)
								or (Pos ('jvc', S) > 0)
								or (Pos ('jvs', S) > 0)
								or (Pos ('jpl', S) > 0)
								or (Pos ('jmi', S) > 0)
								or (Pos ('jhi', S) > 0)
								or (Pos ('jls', S) > 0)
								or (Pos ('jge', S) > 0)
								or (Pos ('jlt', S) > 0)
								or (Pos ('jgt', S) > 0)
								or (Pos ('jle', S) > 0) then begin
								Dest.Delete (LineIndex - 1);
								Dec (LineIndex);
							end;
							S := Strings [I];
						end;
					end;
					Dest.Insert (LineIndex, S);
					Inc (LineIndex);
				end;
			end;
end;

procedure TPatch.LoadFromFile(const FileName: string);
var
	Contents: TStringList;
	CurLine: Integer;
procedure Error(Token: string);
var
	P: Integer;
begin
	Token := Trim (Token);
	P := Pos (' ', Token);
	if P > 0 then
		System.Delete (Token, P, Length (Token));
	if Length (Token) <= 0 then
		Token := 'EOL';
	raise EFileParseError.Create (Format (SParseError, [CurLine + 1, Token]));
end;
procedure ProcessPatch(Patch: TPatch); forward;
function TryPatch(ParentPatch: TPatch): Boolean;
var
	L,
	S: string;
	I,
	P: Integer;
	Patch: TPatch;
	FoundClose: Boolean;
begin
	L := Contents [CurLine];
	P := Pos (PatchToken, UpperCase (L));
	Result := P > 0;
	if Result then begin
		if Length (Trim (Copy (L, 1, P - 1))) > 0 then
			Error (L);
		System.Delete (L, 1, P - 1 + Length (PatchToken));
		Patch := TPatch.Create;
		ParentPatch.GlobalPart.AddObject (PatchToken, Patch);
		L := Trim (L);
		if (Length (L) > 0) and (L [1] = '(') then begin
			System.Delete (L, 1, 1);
			FoundClose := False;
			while not FoundClose do begin
				P := Pos (',', L);
				while P > 0 do begin
					S := Trim (Copy (L, 1, P - 1));
					if Length (S) > 0 then
						ParentPatch.AddObject (S, Patch)
					else
						Error (',');
					System.Delete (L, 1, P);
					P := Pos (',', L);
				end;
				P := Pos (')', L);
				FoundClose := P > 0;
				if P > 0 then begin
					S := Trim (Copy (L, 1, P - 1));
					if Length (S) > 0 then
						ParentPatch.AddObject (S, Patch)
					else
						Error (')');
					System.Delete (L, 1, P);
					P := Pos ('{', L);
					if P > 0 then begin
						S := Trim (Copy (L, 1, P - 1));
						if Length (S) > 0 then
							Error (S);
						System.Delete (L, 1, P);
						if Length (Trim (L)) > 0 then
							Error (L);
					end else begin
						Inc (CurLine);
						if CurLine < Contents.Count then begin
							L := Contents [CurLine];
							P := Pos ('{', L);
							if P > 0 then begin
								S := Trim (Copy (L, 1, P - 1));
								if Length (S) > 0 then
									Error (S);
								System.Delete (L, 1, P);
								if Length (Trim (L)) > 0 then
									Error (L);
							end else
								Error (L);
						end else
							Error ('EOF');
					end;
				end;
				Inc (CurLine);
				if CurLine < Contents.Count then
					L := Contents [CurLine]
				else
					Error ('EOF');
			end;
			ProcessPatch (Patch);
			with Patch do
				for I := 0 to Count - 1 do
					ParentPatch.AddObject (Strings [I], Objects [I]);
		end else
			Error (L);
	end;
end;
procedure ProcessPatch(Patch: TPatch);
begin
	while (CurLine < Contents.Count) and ((Length (Contents [CurLine]) <= 0) or (Contents [CurLine] [1] in [';', '|']) or (Pos ('}', Contents [CurLine]) <= 0)) do begin
		if (Length (Contents [CurLine]) > 0) and (not (Contents [CurLine] [1] in [';', '|'])) and (not TryPatch (Patch)) then
			Patch.GlobalPart.Add (Contents [CurLine]);
		Inc (CurLine);
	end;
	if CurLine < Contents.Count then
		if Length (Trim (Copy (Contents [CurLine], Pos ('}', Contents [CurLine]) + 1, Length (Contents [CurLine])))) > 0 then
			Error (Copy (Contents [CurLine], Pos ('}', Contents [CurLine]) + 1, Length (Contents [CurLine])));
	if (CurLine >= Contents.Count) and (Patch <> Self) then
		Error ('EOF');
end;
begin
	Clear;
	Contents := TStringList.Create;
	try
		Contents.LoadFromFile (FileName);
		CurLine := 0;
		ProcessPatch (Self);
		if CurLine < Contents.Count then
			Error ('}');
	finally
		Contents.Free;
	end;
end;

procedure TPatch.UseKeyWord(const KeyWord: string);
var
	I: Integer;
begin
	if Find (KeyWord, I) then begin
		while (I >= 0) and (I < Count) and (Strings [I] = KeyWord) do
			Dec (I);
		Inc (I);
		while (I >= 0) and (I < Count) and (Strings [I] = KeyWord) do begin
			(Objects [I] as TPatch).Used := True;
			Inc (I);
		end;
	end;
end;

function GetCalcVarSize(ContentLength: LongWord; const Extension: string = ''): LongWord;
begin
	Result := SizeOf (TCalcVarHeader) + ContentLength + SizeOf (TCalcVarFooter);
	if Length (Extension) > 0 then
		Inc (Result, Length (Extension) + 2);
	if Result > MaxCalcAllocBlock then
		Result := 0
end;

function GetTransferFileSize(ContentLength: LongWord; const Extension: string = ''): LongWord;
begin
	Result := GetCalcVarSize (ContentLength, Extension);
	if Result > 0 then
		Inc (Result, SizeOf (TTransferFileHeader) + SizeOf (TTransferFileFooter));
end;

procedure ProduceTransferFile(Output: Pointer; Contents: Pointer; ContentLength: LongWord; CalcDest: TCalcDest; const DestFolder, DestVarName: string; VarLinkType, VarTag: Byte; const Extension: string = '');
const
	Res1: TTFRes2 = ($01, $00);
	Res2: TTFRes6 = ($01, $00, $52, $00, $00, $00);
	Res3: TTFRes3 = ($00, $00, $00);
	Res4: TTFRes6 = ($A5, $5A, $00, $00, $00, $00);
var
	Size,
	CalcSize: LongWord;
	Header: PTransferFileHeader;
	CalcHeader: PCalcVarHeader;
	CalcFooter: PCalcVarFooter;
	Footer: PTransferFileFooter;
	Data: Pointer;
	I,
	Sum: Word;
begin
	Size := GetTransferFileSize (ContentLength, Extension);
	CalcSize := GetCalcVarSize (ContentLength, Extension);
	FillChar (Output^, Size, 0);
	if Size > 0 then begin
		Header := Output;
		CalcHeader := Pointer (@(PBinData(Output) [SizeOf (TTransferFileHeader)]));
		CalcFooter := Pointer (@(PBinData(Output) [Size - SizeOf (TTransferFileFooter) - SizeOf (TCalcVarFooter)]));
		Footer := Pointer (@(PBinData(Output) [Size - SizeOf (TTransferFileFooter)]));
		Data := @(PBinData(Output) [SizeOf (TTransferFileHeader) + SizeOf (TCalcVarHeader)]);
		with Header^ do begin
			if CalcDest = cdTI89 then
				Signature := '**TI89**'
			else
				Signature := '**TI92P*';
			Reserved1 := Res1;
			Reserved2 := Res2;
			Reserved3 := Res3;
			Reserved4 := Res4;
			StrPLCopy (Folder, LowerCase (DestFolder), SizeOf (Folder));
			CharLower (Folder);
			StrPLCopy (VarName, LowerCase (DestVarName), SizeOf (VarName));
			CharLower (VarName);
			WritePCInt (VarLinkType, LinkType);
			WritePCInt (Size, FileSize);
		end;
		WriteCalcInt (CalcSize - SizeOf (TCalcInt2), CalcHeader.DataSize);
		Move (Contents^, Data^, ContentLength);
		if Length (Extension) > 0 then
			StrPCopy (Pointer (@(PBinData(Data) [ContentLength + 1])), Extension);
		WriteCalcInt (VarTag, CalcFooter.Tag);
		Sum := 0;
		for I := 0 to CalcSize - 1 do
			Inc (Sum, Byte (PBinData(CalcHeader) [I]));
		WritePCInt (Sum, Footer.CheckSum);
	end;
end;

end.
