program TiGcc;

uses
	Windows, Classes, SysUtils,
	ParsingUnit in '..\ParsingUnit.pas',
	ObjectFileUnit in '..\ObjectFileUnit.pas';

const
	SErrorPrefix = 'TIGCC.EXE: ';
	SFileNotFound = '%s: No such file or directory';
	SNoFiles = 'No input files';
	SRunFailed = 'Program execution failed';
	SProgNotFound = 'Installation error: Cannot find %s';
	SQuillNotFound = 'File ''Quill.drv'' not found in Bin or Include directory';
	SDLLTooLarge = 'DLL size of %ld bytes is too large';
	SCompressedTooLarge = 'Compressed size of %ld bytes is too large, unexpectedly';

procedure Error(const Msg: string = '');
var
	TempStr: array [0..511] of Char;
	WriteResult: Cardinal;
begin
	if Length (Msg) > 0 then begin
		StrPLCopy (TempStr, SErrorPrefix + Msg + #13#10, 511);
		WriteFile (GetStdHandle (Std_Error_Handle), TempStr, Length (Msg) + Length (SErrorPrefix) + 2, WriteResult, nil);
	end;
end;

procedure Fatal(const Msg: string = '');
begin
	raise Exception.Create (Msg);
end;

function LastPos(const Substr, S: string): Integer;
var
	I: Integer;
begin
	Result := 0;
	for I := Length (S) - Length (Substr) + 1 downto 1 do
		if Copy (S, I, Length (SubStr)) = Substr then begin
			Result := I;
			Break;
		end;
end;

function GetShortFileName(FileName: string): string;
var
	F: TSearchRec;
	Found: Boolean;
begin
	Result := '';
	while Length (FileName) > 2 do begin
		Found := FindFirst (FileName, faAnyFile, F) = 0;
		while Found and ((F.Name = '') or (F.Name = '.') or (F.Name = '..')) do
			Found := FindNext (F) = 0;
		if Found then begin
			if F.FindData.cAlternateFileName <> '' then
				Result := F.FindData.cAlternateFileName + '\' + Result
			else if Length (F.Name) > 0 then
				Result := F.Name + '\' + Result
			else
				Result := ExtractFileName (FileName) + '\' + Result;
		end else
			Result := ExtractFileName (FileName) + '\' + Result;
		FindClose (F);
		Delete (FileName, LastPos ('\', FileName), Length (FileName));
	end;
	Result := FileName + '\' + Copy (Result, 1, Length (Result) - 1);
end;

function Enquote(const S: string): string;
begin
	if Pos (' ', S) > 0 then
		Result := '"' + S + '"'
	else
		Result := S;
end;

procedure InsertOptionString(var Dest: string; Options: string);
var
	P: Integer;
begin
	Options := Options + ',';
	while Length (Options) > 0 do begin
		P := Pos (',', Options);
		Dest := Dest + ' ' + Enquote (Copy (Options, 1, P - 1));
		Delete (Options, 1, P);
	end;
end;

const
	NewLine: array [0..1] of Char = (#13, #10);

procedure InsertNewLine;
var
	C: Cardinal;
begin
	WriteFile (GetStdHandle (Std_Output_Handle), NewLine, SizeOf (NewLine), C, nil);
end;

var
	Verbose: Boolean = False;
	Quiet: Boolean = False;

procedure Execute(const CommandLine: string);
var
	TempStr: array [0..255] of Char;
	WriteResult: Cardinal;
	StartupInfo: TStartupInfo;
	ProcessInfo: TProcessInformation;
begin
	if FileExists (Copy (CommandLine, 1, Pos (' ', CommandLine))) then begin
		if Verbose then begin
			StrPLCopy (TempStr, CommandLine + #13#10, 255);
			WriteFile (GetStdHandle (Std_Output_Handle), TempStr, Length (CommandLine) + 2, WriteResult, nil);
		end;
		FillChar (StartupInfo, SizeOf (StartupInfo), 0);
		with StartupInfo do begin
			cb := SizeOf (StartupInfo);
			lpTitle := PChar ('TIGCC');
		end;
		if CreateProcess (nil, PChar (CommandLine), nil, nil, True, 0, nil, nil, StartupInfo, ProcessInfo) then begin
			WaitForSingleObject (ProcessInfo.hProcess, Infinite);
			GetExitCodeProcess (ProcessInfo.hProcess, WriteResult);
			CloseHandle (ProcessInfo.hProcess);
			CloseHandle (ProcessInfo.hThread);
			if WriteResult <> 0 then
				Fatal;
		end else
			Fatal (SRunFailed);
	end else
		Fatal (Format (SProgNotFound, [Copy (CommandLine, 1, Pos (' ', CommandLine))]));
end;

var
	Assemble: Boolean = True;
	Link: Boolean = True;
	Archive: Boolean = False;
	DebugInfo: Boolean = False;
	OutputBin: Boolean = False;
	StdLib: Boolean = True;
	Pack: Boolean = False;
	KeepObjectFiles: Boolean = False;
	SaveTemps: Boolean = False;
	FirstFile: string = '';
	DestFile: string = '';
	LdFiles: string = '';
	PackVar: string = '';
	LdFileCount: Integer = 0;
	GCCFiles,
	A68kFiles: TStringList;
	DebugFiles: TStringList = nil;
	A68kDest: Boolean = False;
	ProgPath,
	GCCLine,
	AsLine,
	A68kLine,
	ObjCopyLine,
	Obj2TILine,
	LdLine,
	ArLine,
	PackLine,
	PackObj2TILine: string;
	I: Integer;
	S,
	T: string;
	L,
	DelFiles: TStringList;
	PatchLib,
	PatchMainLib: TPatch;
	PatchIncluded: Boolean;
	OutputSize: LongWord;
	OutputStream: TMemoryStream;

begin
	ParsingUnit.ErrorMessageProc := Error;

	DelFiles := TStringList.Create;
	GCCFiles := TStringList.Create;
	A68kFiles := TStringList.Create;

	try
		try

			if ParamCount <= 0 then
				Fatal (SNoFiles);

			// Basic command lines
			ProgPath := ExtractFilePath (GetShortFileName (ParamStr (0)));
			GCCLine := ProgPath + 'BIN\GCC.EXE -B' + ProgPath + 'BIN\ -I ' + ProgPath + 'INCLUDE\C\';
			AsLine := ProgPath + 'BIN\AS.EXE -mc68000 -I ' + ProgPath + 'INCLUDE\S\';
			A68kLine := ProgPath + 'BIN\A68K.EXE -i' + ProgPath + 'INCLUDE\ASM\ -g -t';
			ObjCopyLine := ProgPath + 'BIN\OBJCOPY.EXE -O coff-m68k';
			LdLine := ProgPath + 'BIN\LD.EXE -r -e 0 -o tempfile.o';
			ArLine := ProgPath + 'BIN\AR.EXE -rc';
			Obj2TILine := ProgPath + 'BIN\OBJ2TI.EXE';
			PackLine := ProgPath + 'BIN\PACK.EXE %s program.pck';
			PackObj2TILine := Obj2TILine + ' pstarter.o';

			// Parsing of command line arguments
			I := 1;
			while I <= ParamCount do begin
				S := ParamStr (I);
				if (Length (S) >= 2) and (S [1] = '-') then begin
					if S = '-bsr' then
						// ignore for compatibility
					else if (S = '-outputbin') or (S = '--outputbin') then
						OutputBin := True
					else if (S = '-standalone') or (S = '--standalone') then
						StdLib := False
					else if S = '-ar' then begin
						Archive := True;
						StdLib := False;
					end else if (S = '-keep') or (S = '--keep') then
						KeepObjectFiles := True
					else if (S = '-save-temps') or (S = '--save-temps') then begin
						KeepObjectFiles := True;
						SaveTemps := True;
						Insert (' ' + Enquote (S), GCCLine, Length (GCCLine) + 1);
					end else if S = '-include' then begin
						if I < ParamCount then begin
							Inc (I);
							Insert (' ' + Enquote (S) + ' ' + Enquote (ParamStr (I)), GCCLine, Length (GCCLine) + 1);
						end;
					end else if S = '-pack' then begin
						Pack := True;
						if I < ParamCount then begin
							Inc (I);
							PackVar := LowerCase (Copy (ParamStr (I), 1, 8));
						end else begin
							Randomize;
							PackVar := 'prog' + IntToStr (Random (9999));
						end;
						DelFiles.Add ('program.pck');
					end else if S = '-quill' then begin
						if FileExists (ProgPath + 'BIN\Quill.drv') then
							Insert (' -Os -include "' + ProgPath + 'BIN\Quill.drv" -x c', GCCLine, Length (GCCLine) + 1)
						else if FileExists (ProgPath + 'INCLUDE\C\Quill.drv') then
							Insert (' -Os -include "' + ProgPath + 'INCLUDE\C\Quill.drv" -x c', GCCLine, Length (GCCLine) + 1)
						else if FileExists (ProgPath + 'INCLUDE\QUILL\Quill.drv') then
							Insert (' -Os -include "' + ProgPath + 'INCLUDE\QUILL\Quill.drv" -x c', GCCLine, Length (GCCLine) + 1)
						else
							Fatal (SQuillNotFound);
					end else if Copy (S, 1, 4) = '-Wa,' then
						InsertOptionString (AsLine, Copy (S, 5, Length (S)))
					else if Copy (S, 1, 4) = '-WA,' then
						InsertOptionString (A68kLine, Copy (S, 5, Length (S)))
					else if Copy (S, 1, 4) = '-Wl,' then
						InsertOptionString (LdLine, Copy (S, 5, Length (S)))
					else if Copy (S, 1, 2) = '-v' then begin
						Verbose := True;
						if S <> '-v0' then begin
							Quiet := False;
							Insert (' ' + Enquote (S), GCCLine, Length (GCCLine) + 1);
							Insert (' -v', PackLine, Length (PackLine) + 1);
						end;
					end else if Length (S) = 2 then begin
						case S [2] of
							'E', 'S': begin
								Link := False;
								Assemble := False;
								if S [2] = 'E' then
									Insert (' ' + Enquote (S), GCCLine, Length (GCCLine) + 1);
							end;
							'c':
								Link := False;
							'q':
								Quiet := True;
							'g':
								DebugInfo := True;
							'x':
								if I < ParamCount then begin
									Inc (I);
									Insert (' ' + Enquote (ParamStr (I)), GCCLine, Length (GCCLine) + 1);
								end;
							'o':
								if I < ParamCount then begin
									Inc (I);
									DestFile := ParamStr (I);
								end;
							'8', '9':
								Insert (' ' + Enquote (S), Obj2TILine, Length (Obj2TILine) + 1);
							else
								Insert (' ' + Enquote (S), GCCLine, Length (GCCLine) + 1);
						end;
					end else begin
						if S = '--help' then begin
							Assemble := False;
							Link := False;
						end;
						if S [2] = 'g' then
							KeepObjectFiles := True;
						Insert (' ' + Enquote (S), GCCLine, Length (GCCLine) + 1);
					end;
				end else begin
					if LowerCase (ExtractFileExt (S)) = '.a' then begin
						if (not FileExists (S)) and FileExists (ProgPath + 'LIB\' + S) then
							S := ProgPath + 'LIB\' + S;
						Insert (' ' + Enquote (S), LdFiles, Length (LdFiles) + 1);
						Inc (LdFileCount);
					end else begin
						if FileExists (S) then begin
							Insert (' ' + Enquote (ChangeFileExt (S, '.o')), LdFiles, Length (LdFiles) + 1);
							Inc (LdFileCount);
							if Length (FirstFile) <= 0 then
								FirstFile := Copy (S, 1, Pos (ExtractFileExt (S), S) - 1);
							if LowerCase (ExtractFileExt (S)) <> '.o' then begin
								if LowerCase (ExtractFileExt (S)) = '.asm' then
									A68kFiles.Add (S)
								else begin
									Insert (' ' + Enquote (S), GCCLine, Length (GCCLine) + 1);
									GCCFiles.Add (S);
								end;
								if Link then
									DelFiles.Add (ChangeFileExt (S, '.o'));
							end;
						end else
							Error (Format (SFileNotFound, [S]));
					end;
				end;
				Inc (I);
			end;
			if StdLib and (LdFileCount > 0) then begin
				Insert (' ' + ProgPath + 'LIB\tigcc.a', LdFiles, Length (LdFiles) + 1);
				Inc (LdFileCount);
			end;
			if DebugInfo or KeepObjectFiles then
				with DelFiles do
					for I := Count - 1 downto 0 do
						if LowerCase (ExtractFileExt (Strings [I])) = '.o' then
							Delete (I);

			// Execution of GCC.exe (compiling and assembling)
			if (GCCFiles.Count > 0) or (Pos ('--', GCCLine) > 0) then begin
				if (Pos (' -S', GCCLine) <= 0) and (Pos (' -E', GCCLine) <= 0) then
					Insert (' -S', GCCLine, Length (GCCLine) + 1);
				if DebugInfo then
					Insert (' -gcoff -mcoff-abslines', GCCLine, Length (GCCLine) + 1);
				if (GCCFiles.Count = 1) and (A68kFiles.Count <= 0) and (Length (DestFile) > 0) and (not Assemble) then
					Insert (' -o ' + Enquote (DestFile), GCCLine, Length (GCCLine) + 1);
				Execute (GCCLine);
				PatchLib := TPatch.Create;
				PatchMainLib := TPatch.Create;
				try
					PatchLib.Used := True;
					try
						if FileExists (ProgPath + 'LIB\TIPatch.lib') then
							PatchLib.LoadFromFile (ProgPath + 'LIB\TIPatch.lib');
					except
						on E: EFileParseError do
							Fatal ('TIPATCH.LIB ' + E.Message);
						else
							raise;
					end;
					PatchMainLib.Used := True;
					try
						if FileExists (ProgPath + 'LIB\TIPatchMain.lib') then
							PatchMainLib.LoadFromFile (ProgPath + 'LIB\TIPatchMain.lib');
					except
						on E: EFileParseError do
							Fatal ('TIPATCHMAIN.LIB ' + E.Message);
						else
							raise;
					end;
					for I := 0 to GCCFiles.Count - 1 do begin
						S := GCCFiles.Strings [I];
						T := ExtractFileName (ChangeFileExt (S, '.s'));
						if FileExists (T) then begin
							if LowerCase (ExtractFileExt (S)) <> '.s' then begin
								if Assemble then begin
									if DebugInfo then begin
										if not Assigned (DebugFiles) then
											DebugFiles := TStringList.Create;
										DebugFiles.Add (S);
									end else
										DelFiles.Add (T);
								end;
								L := TStringList.Create;
								with L do try
									LoadFromFile (T);
									PatchLib.DeselectAll;
									PatchMainLib.DeselectAll;
									ParseSFile (L, PatchLib, PatchMainLib, PatchIncluded);
									if DebugInfo and (not Assemble) then
										ParseDebugSFileDisk (L, S);
									SaveToFile (T);
								finally
									Free;
								end;
							end;
							if Assemble then begin
								if (not Link) and (Length (DestFile) > 0) then
									Execute (AsLine + ' "' + T + '" -o ' + DestFile)
								else
									Execute (AsLine + ' "' + T + '" -o "' + ChangeFileExt (S, '.o') + '"');
							end;
						end;
					end;
				finally
					PatchMainLib.Free;
					PatchLib.Free;
				end;
				if Link and Verbose then
					InsertNewLine;
			end;

			// Execution of A68k.exe (assembling)
			if A68kFiles.Count > 0 then begin
				if DebugInfo then
					Insert (' -d', A68kLine, Length (A68kLine) + 1);
				if Quiet then
					Insert (' -q', A68kLine, Length (A68kLine) + 1);
				if (A68kFiles.Count = 1) and (GCCFiles.Count <= 0) and (Length (DestFile) > 0) and (not Link) then begin
					Insert (' ' + Enquote ('-o' + DestFile), A68kLine, Length (A68kLine) + 1);
					A68kDest := True;
				end;
				while A68kFiles.Count > 0 do begin
					Execute (A68kLine + ' "' + A68kFiles.Strings [0] + '"');
					if A68kDest then
						Execute (ObjCopyLine + ' ' + DestFile)
					else
						Execute (ObjCopyLine + ' "' + ChangeFileExt (A68kFiles.Strings [0], '.o') + '"');
					A68kFiles.Delete (0);
					if Link and not Quiet then begin
						InsertNewLine;
						InsertNewLine;
					end;
				end;
			end;

			// Execution of Ld.exe and Obj2TI.exe (linking), or Ar.exe (archiving)
			if Link and (LdFileCount > 0) then begin
				if Length (DestFile) <= 0 then
					DestFile := FirstFile;
				if Archive then begin
					if Pos ('.', DestFile) <= 0 then
						DestFile := DestFile + '.a';
					Insert (' ' + Enquote (DestFile), ArLine, Length (ArLine) + 1);
					Insert (LdFiles, ArLine, Length (ArLine) + 1);
					Execute (ArLine);
				end else begin
					Insert (LdFiles, LdLine, Length (LdLine) + 1);
					if not KeepObjectFiles then
						DelFiles.Add ('tempfile.o');
					Execute (LdLine);
					Insert (' tempfile.o', Obj2TILine, Length (Obj2TILine) + 1);
					if Assigned (DebugFiles) then begin
						ParseDebugFiles (DebugFiles, 'tempfile.o');
						DebugFiles.Free;
					end;
					if Verbose then
						InsertNewLine;
					if OutputBin or Pack then
						Insert (' --outputbin', Obj2TILine, Length (Obj2TILine) + 1);
					if Quiet then
						Insert (' -q', Obj2TILine, Length (Obj2TILine) + 1);
					if Length (DestFile) > 0 then
						Insert (' -o ' + Enquote (DestFile), Obj2TILine, Length (Obj2TILine) + 1);
					Execute (Obj2TILine);
					if FileExists (DestFile + '.ndl') and (Length (DestFile) > 0) then begin
						DelFiles.Add (DestFile + '.ndl');
						with TMemoryStream.Create do try
							LoadFromFile (DestFile + '.ndl');
							OutputSize := GetTransferFileSize (Size, 'DLL');
							if OutputSize > 0 then begin
								OutputStream := TMemoryStream.Create;
								try
									OutputStream.Size := OutputSize;
									ProduceTransferFile (OutputStream.Memory, Memory, Size, cdTI89, 'main', ExtractFileName (DestFile), $1C, $F8, 'DLL');
									OutputStream.SaveToFile (DestFile + '.89y');
									ProduceTransferFile (OutputStream.Memory, Memory, Size, cdTI92Plus, 'main', ExtractFileName (DestFile), $1C, $F8, 'DLL');
									OutputStream.SaveToFile (DestFile + '.9xy');
								finally
									OutputStream.Free;
								end;
							end else
								Fatal (Format (SDLLTooLarge, [Size]));
						finally
							Free;
						end;
					end;
				end;
			end;

			// Execution of Pack.exe (packing)
			if Pack and Link and (not Archive) and (Length (DestFile) > 0) then begin
				if FileExists (DestFile + '.89y') then try
					DeleteFile (DestFile + '.89y');
				except end;
				if FileExists (DestFile + '.9xy') then try
					DeleteFile (DestFile + '.9xy');
				except end;
				if Quiet then
					Insert (' -quiet', PackLine, Length (PackLine) + 1);
				DelFiles.Add (DestFile + '.bin');
				Execute (Format (PackLine, [DestFile + '.bin']));
				with TMemoryStream.Create do try
					LoadFromFile ('program.pck');
					OutputSize := GetTransferFileSize (Size, 'ppg');
					if OutputSize > 0 then begin
						OutputStream := TMemoryStream.Create;
						try
							OutputStream.Size := OutputSize;
							ProduceTransferFile (OutputStream.Memory, Memory, Size, cdTI89, 'main', PackVar, $1C, $F8, 'ppg');
							OutputStream.SaveToFile (DestFile + '.89y');
							ProduceTransferFile (OutputStream.Memory, Memory, Size, cdTI92Plus, 'main', PackVar, $1C, $F8, 'ppg');
							OutputStream.SaveToFile (DestFile + '.9xy');
						finally
							OutputStream.Free;
						end;
					end else
						Fatal (Format (SCompressedTooLarge, [Size]));
				finally
					Free;
				end;
				DelFiles.Add ('pstarter.o');
				ParsePStarter (ProgPath + 'LIB\pstarter.o', 'pstarter.o', PackVar);
				Execute (PackObj2TILine + ' -o ' + DestFile);
			end;

			if not SaveTemps then
				with DelFiles do
					for I := 0 to Count - 1 do
						if FileExists (Strings [I]) then try
							DeleteFile (Strings [I]);
						except end;

		finally
			A68kFiles.Free;
			GCCFiles.Free;
			DelFiles.Free;
		end;

	except
		on E: Exception do begin
			if not (E is EAbort) then
				Error (E.Message);
			Halt (1);
		end;
	end;
end.
