/*
   tprbuilder - tprbuilder is a MAKE program for TIGCC projects (.tpr files)
   
   Copyright (C) 2002 Romain Livin
   Copyright (C) 2002 Kevin Kofler

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <getopt.h>
#include "system.h"

#include "../shared/misc.h"
#include "../shared/misc.c"

#include "tprbuilder.h"

#define EXIT_FAILURE 1

/* The name the program was run with, stripped of any leading path. */
char *program_name;

/* Option flags and variables */
static struct option const long_options[] =
{
    {"help", no_argument, 0, 'h'},
    {"version", no_argument, 0, 'V'},
    {"clean", no_argument, 0, 'c'},
    {"quiet", no_argument, 0, 'q'},
    {"verbose", no_argument, 0, 'v'},
    {NULL, 0, NULL, 0}
};

/* global arrays */
char **c_files    = NULL;
char **s_files    = NULL;
char **h_files    = NULL;
char **a_files    = NULL;
char **asm_files  = NULL;

/* global vars */
char *tigcc_base = NULL;
int c_file_count, s_file_count, m_file_count, h_file_count, asm_file_count, a_file_count;
Settings settings = { 0 };

int clean = 0;
int verbose = 0;
int quiet = 0;


/*
   Our main function
*/
int main(int argc, char **argv)
{
    int i;
    FILE *f;
    char *filename;
    char buffer[256];
    int l = 0;

    /* GNU stuffs */
    program_name = argv[0];
    
    i = decode_switches (argc, argv);
    
    /* Do some checkings */
    if (argc < 2) {
	fprintf(stderr, "tprbuilder: no input files\n");
	exit(0);
    }
    
#ifdef __WIN32__
/*    tigcc_base = alloca(strlen(argv[0])+1);
    {
        char *p = strrchr(tigcc_base,'\\');
        if (p) *p = 0; else {
            fprintf(stderr, "Fatal error: Could not detect TIGCC path\n");
            exit(-1);
        }
    } */
#else
    /* Retrieve environment variable */
    if ((tigcc_base = getenv("TIGCC")) == NULL) {
	fprintf(stderr, "Fatal error: TIGCC is not defined in the environment. TIGCC must be defined before tigcc can run.\nFor (ba)sh, try: export TIGCC=path/to/tigcc\nFor (t)csh, try: setenv TIGCC /path/to/tigcc\n");
	exit(-1);
    }
#endif
    
    /* Register exit callback */
    if (atexit(safe_exit) != 0) {
	fprintf(stderr, "Fatal error: unable to register safe exit callback\n");
	exit(-1);
    }

    /* Parse for files */
    if (i < argc)
    {
	if(strstr(argv[i], ".tpr")) {
	    parse_file(argv[i]);
	}
	else {
	    usage(EXIT_FAILURE);
	    printf ("non-option ARGV-elements: ");
	    while (i < argc)
		printf ("%s ", argv[i++]);
	}
    }

    /* Build files */
    if(!clean) {
	build_files();
    }
    
    /* Clean-up (if enabled) */
    if(clean) {
	clean_files();
    }

    return 0;
}

/* 
   Set all the option flags according to the switches specified.
   Return the index of the first non-option argument.  
*/
static int decode_switches (int argc, char **argv)
{
  int c;

  while ((c = getopt_long (argc, argv, 
			   "h"	/* help */
			   "V",	/* version */
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 'V':
	    printf ("tprbuilder 1.0.6\n");
	    exit (0);
	    break;
	case 'h':
	    usage (0);
	    exit(0);
	    break;
	case 'c':
	    clean = 1;
	    break;
	case 'q':
	    quiet = 1;
	    break;
	case 'v':
	    verbose = 1;
	    break;
	    
	case '?':
	    break;

	default:
	    usage (EXIT_FAILURE);
	    break;
	}
    }

  return optind;
}

/*
  Display usage options
*/
static void usage (int status)
{
    printf (_("%s - \
tprbuilder is a MAKE program for TIGCC projects (.tpr files)\n"), program_name);
    printf (_("Usage: %s [OPTION]... [FILE]...\n"), program_name);
    printf (_("\
Options:\n\
  -h, --help                 display this help and exit\n\
  -V, --version              output version information and exit\n\
  -v, --verbose              enable verbose mode (make tigcc verbose)\n\
  -q, --quiet                enable quiet mode (display nothing)\n\
  --clean                    do a cleanup of giles\n\
"));
    exit (status);
}

/* 
   Destroy allocated resources (exit callback)
 */
void safe_exit(void)
{
    free(settings.pack_name);
    free(settings.prj_name);
    free(settings.cc_switches);
    free(settings.as_switches);
    free(settings.ld_switches);
    free(settings.cv_switches);
    free(settings.a68k_switches);
    free(settings.post_build);
    free(settings.cmdline);
    
    dyn_file_array_free(c_files);
    dyn_file_array_free(s_files);
    dyn_file_array_free(h_files);
    dyn_file_array_free(asm_files);
    dyn_file_array_free(a_files);
}

/* 
   Remove/unlink a file
*/
int delete(char *filename)
{
    if (!quiet) {
        fprintf(stderr, "tprbuilder: removing %s\n", filename);
    }

    tail_long_filename(filename);
    remove(filename);

    return 0;
}

/* 
   Execute a program by calling the shell
*/
int execute(char *cmdline)
{
    if (!quiet) {
        fprintf(stderr, "tprbuilder: %s\n", cmdline);
    }

    if(system(cmdline) != 0) {
	exit(0);
    }

    return 0;
}

/*
  Execute tigcc
*/
int execute_tigcc(char *filename, char *args)
{
    char *cmdline, *quill_flag = "";

    /* We need to do this here because Win32 tigcc.exe only understands -quill
       at the beginning of the command line. */
    if(settings.quill == TRUE) quill_flag = "-quill ";

#ifdef __WIN32__
    cmdline = calloc(strlen(filename) + strlen(args) + 32, sizeof(char));
    sprintf(cmdline, "tigcc.exe %s%s %s", quill_flag, filename, args);
#else
    cmdline = calloc(strlen(tigcc_base) + strlen(filename) + strlen(args) +
		     32, sizeof(char));
    sprintf(cmdline, "%s/bin/tigcc %s%s %s", tigcc_base, quill_flag, filename, args);
#endif
    if(verbose)
      strcat(cmdline, " --verbose");
    execute(cmdline);
    free(cmdline);
}


/* 
   Strip CR and/or LF terminators 
*/
char* strip(char *str)
{
    int len = strlen(str);

    if(len > 0)
    	if( (str[len-1] == '\r') || (str[len-1] == '\n') )
			str[len-1] = '\0';
    
    if(len > 1)
    	if( (str[len-2] == '\r') || (str[len-2] == '\n') )
	    str[len-2] = '\0';
    
    return str;
}

/* 
   Print an error msg 
*/
void stop (int line)
{
    fprintf(stderr, "Configuration file error at line %i.\n", line);
}

/* 
   Read a line from file and do a clean-up 
*/
int read_line(FILE *f, char *buffer, int *l)
{
    strcpy(buffer, "");
    if(feof(f)) return EOF;
    fgets(buffer, 256, f);
    strip(buffer);
    (*l)++;
    while( (buffer[0] == '\r') || (buffer[0] == '\n') || (buffer[0] == '\0') )
    {
      	if(feof(f)) return EOF;
      	fgets(buffer, 256, f);
      	strip(buffer);
      	(*l)++;
    }
    
    return 0;
}


/* 
   Same as above but this one can extract a number such as scanf 
*/
char *find_param_str(char *s, const char*t, int *i)
{
    char *p;
    char buf[256];
    int ret = 0;
    char arglist[256] = "";
    
    p = (char *)strstr(t, "%i");
    if(p == NULL) return NULL;
    
    strcat(arglist, t);
    strcat(arglist, "%s");
    
    ret = sscanf(s, arglist, i, buf);
    if(ret < 2) return NULL;
    
    p = (char *)strstr(s, buf);
    return p;
}


/* 
   Parse a TPR file, fill the Setting structure and the filename arrays 
*/
int parse_file(const char *filename)
{
    FILE *f;
    char buffer[256];
    int l = 0;
    SectionType stype = SECTION_NONE;
    
    f = fopen(filename, "rt");
    if(f == NULL) {
	fprintf(stderr, "Unable to open this file: <%s>\n", filename);
	exit(-1);
    }
    
    while(!feof(f)) 
    {
  	char *p;
  	
  	// Get a line from file
  	if(read_line(f, buffer, &l) == EOF) break;
	
	// Search for sections
	if( (p=find_str(buffer, TOKEN_SETTINGS)) )
	{
	    stype = SECTION_SETTINGS;
	    continue;
	}
	
	if( (p=find_str(buffer, TOKEN_FILES)) )
	{
	    stype = SECTION_FILES;
	    continue;
	}
	
	// Keywords in the [Settings] section
	if(stype == SECTION_SETTINGS) {

	    if( (p=find_str(buffer, TOKEN_ARCHIVE)) )
	    {
		if(!strcmp(p, "0")) settings.archive = FALSE;
		else if(!strcmp(p, "1")) settings.archive = TRUE;
		else stop(l);
		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_PACK)) )
	    {
		if(!strcmp(p, "0")) settings.pack = FALSE;
		else if(!strcmp(p, "1")) settings.pack = TRUE;
		else stop(l);
		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_PACKNAME)) )
	    {
      		settings.pack_name = strdup(p);
      		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_PRJNAME)) )
	    {
      		settings.prj_name = strdup(p);
      		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_CC)) )
	    {
      		settings.cc_switches = strdup(p);
      		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_A68K)) )
	    {
      		settings.a68k_switches = strdup(p);
      		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_CV)) )
	    {
      		settings.cv_switches = strdup(p);
      		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_LD)) )
	    {
      		settings.ld_switches = strdup(p);
      		continue;
	    }

	    else if( (p=find_str(buffer, TOKEN_AS)) )
	    {
      		settings.as_switches = strdup(p);
      		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_BSRPATCH)) )
	    {
		if(!strcmp(p, "0")) settings.bsr_patch = FALSE;
		else if(!strcmp(p, "1")) settings.bsr_patch = TRUE;
		else stop(l);
		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_DEBUGINFO)) )
	    {
		if(!strcmp(p, "0")) settings.debug_info = FALSE;
		else if(!strcmp(p, "1")) settings.debug_info = TRUE;
		else stop(l);
		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_STDLIB)) )
	    {
		if(!strcmp(p, "0")) settings.std_lib = FALSE;
		else if(!strcmp(p, "1")) settings.std_lib = TRUE;
		else stop(l);
		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_CMDLINE)) )
	    {
      		settings.cmdline = strdup(p);
      		continue;
	    }
	    
	    else if( (p=find_str(buffer, TOKEN_POSTBUILD)) )
	    {
		settings.post_build = strdup(p);
		continue;
	    }
	    else stop(l);
	}
	
	// Token in the [Included Files] section
	if(stype = SECTION_FILES)
	{
	    int v;

	    if( (p=find_param_str(buffer, TOKEN_CFILE, &v)) )
	    {
		char *s = encapsulate_long_filename(p);
       		dyn_file_array_add(s, &c_files, &c_file_count);
		free(s);

       		continue;
	    }

	    else if( (p=find_param_str(buffer, TOKEN_SFILE, &v)) )
	    {
		char *s = encapsulate_long_filename(p);
       		dyn_file_array_add(s, &s_files, &s_file_count);
		free(s);

       		continue;
	    }

	    else if( (p=find_param_str(buffer, TOKEN_HFILE, &v)) )
	    {
		char *s = encapsulate_long_filename(p);
       		dyn_file_array_add(s, &h_files, &h_file_count);
		free(s);

		continue;
	    }

	    else if( (p=find_param_str(buffer, TOKEN_ASMFILE, &v)) )
            {
		char *s = encapsulate_long_filename(p);
                dyn_file_array_add(s, &asm_files, &asm_file_count);
		free(s);

		continue;
            }

	    else if( (p=find_param_str(buffer, TOKEN_AFILE, &v)) )
            {
		char *s = encapsulate_long_filename(p);
                dyn_file_array_add(s, &a_files, &a_file_count);
		free(s);

		continue;
            }

	    else if( (p=find_param_str(buffer, TOKEN_TFILE, &v)) )
            { // text file: do nothing
		continue;
	    }

	    else if( (p=find_param_str(buffer, TOKEN_QLLFILE, &v)) )
            { // Quill file: treat like C file and add -quill flag
		char *s = encapsulate_long_filename(p);
       		dyn_file_array_add(s, &c_files, &c_file_count);
		free(s);

		settings.quill = TRUE; // -quill flag needed

       		continue;

	    }

	    else stop(l);
	}

    }
    fclose(f);

    return 0;
}

/*
   Build command line arguments (Settings section)
*/
int process_settings(char *args)
{
    // Process settings
    strcpy(args, " ");
    //strcat(args, " -v");

    if(settings.archive == TRUE) {
	strcat(args, " -ar");
    }

    if(settings.pack == TRUE) {
	strcat(args, " -pack ");
	strcat(args, settings.pack_name);
    }
    
    if(settings.bsr_patch == TRUE) {
	strcat(args, " -bsr");
    }
    
    if(settings.debug_info == TRUE) {
	strcat(args, " -g");
    }

    if(settings.std_lib == FALSE) {
	strcat(args, " -standalone");
    }

    if(settings.prj_name) {
	strcat(args, " -o ");
	strcat(args, settings.prj_name);
    }

    if(settings.cc_switches) {
	strcat(args, " ");
	strcat(args, settings.cc_switches);
    }

    if(settings.as_switches) {
	size_t i;
	strcat(args, " -Wa,");
	for (i=0;i<strlen(settings.as_switches);i++) if (settings.as_switches[i]==' ') settings.as_switches[i]=',';
	strcat(args, settings.as_switches);
    }

    if(settings.ld_switches) {
	size_t i;
	strcat(args, " -Wl,");
	for (i=0;i<strlen(settings.ld_switches);i++) if (settings.ld_switches[i]==' ') settings.ld_switches[i]=',';
	strcat(args, settings.ld_switches);
    }

    if(settings.cv_switches) {
	strcat(args, " ");
	strcat(args, settings.cv_switches);
    }

    if(settings.a68k_switches) {
	size_t i;
	strcat(args, " -WA,");
	for (i=0;i<strlen(settings.a68k_switches);i++) {
	char *p=settings.a68k_switches+i;
	if (*p==',') *p=';'; else if (*p==' ') *p=',';
	}
	strcat(args, settings.a68k_switches);
    }
/* used for what ? , maybe remote control */
/*
    if(settings.cmdline) {
	strcat(args, " ");
	strcat(args, settings.cmdline);
	}*/

    return 0;
}

/*
   Build command line files (Included Files section)
*/
int process_files(char *files, char *o_files)
{
    int i, count;
    char tmpfile[256];

    /* Process .c files */
    dyn_file_array_size(c_files, &count);    
    for(i=0; i<count; i++) {
	    strcat(files, c_files[i]);
	    strcat(files, " ");
	    
	    strcpy(tmpfile, c_files[i]);
	    change_extension(tmpfile, ".o");
	    strcat(o_files, tmpfile);
	    strcat(o_files, " ");
    }

    /* Process .s files */
    dyn_file_array_size(s_files, &count);    
    for(i=0; i<count; i++) {
	    strcat(files, s_files[i]);
	    strcat(files, " ");

	    strcpy(tmpfile, s_files[i]);
	    change_extension(tmpfile, ".o");
	    strcat(o_files, tmpfile);
	    strcat(o_files, " ");
    }

    /* Process .asm files */
    dyn_file_array_size(asm_files, &count);
    for(i=0; i<count; i++) {
	    strcat(files, asm_files[i]);
	    strcat(files, " ");

	    strcpy(tmpfile, asm_files[i]);
	    change_extension(tmpfile, ".o");
	    strcat(o_files, tmpfile);
	    strcat(o_files, " ");
    }

    /* Process .a files */
    dyn_file_array_size(a_files, &count);
    for(i=0; i<count; i++) {
	    strcat(files, a_files[i]);
	    strcat(files, " ");
    }

    return 0;
}

/*
  Call tigcc for making files
*/
int build_files(void)
{
    char options[256] = "";
    char files[65536] = "";
    char o_files[65536] = "";
    int i, count;

    process_settings(options);
    process_files(files, o_files);

    execute_tigcc(files, options);

#if 0
    if(settings.archive == TRUE) {
	char cmdline[256];

	sprintf(cmdline, "%s/bin/%sar -r %s.a %s", tigcc_base, BIN_BASE,
		settings.prj_name , o_files);
	execute(cmdline);
    }
#endif
}


/*
  Remove files generated by tigcc
*/
int clean_files(void)
{
    char tmpfile[256];
    int i, count;

    /* Process .c files */
    dyn_file_array_size(c_files, &count);    
    for(i=0; i<count; i++) {
	strcpy(tmpfile, c_files[i]);
	change_extension(tmpfile, ".s");
	delete(tmpfile);

	strcpy(tmpfile, c_files[i]);
	change_extension(tmpfile, ".o");
	delete(tmpfile);
    }

    /* Process .s files */
    dyn_file_array_size(s_files, &count);    
    for(i=0; i<count; i++) {
	strcpy(tmpfile, s_files[i]);
	change_extension(tmpfile, ".o");
	delete(tmpfile);
    }

    /* Process .asm files */
    dyn_file_array_size(asm_files, &count);    
    for(i=0; i<count; i++) {
	strcpy(tmpfile, asm_files[i]);
	change_extension(tmpfile, ".o");
	delete(tmpfile);
    }

    /* Remove executables */
    strcpy(tmpfile, settings.prj_name);
    strcat(tmpfile, ".89z");
    delete(tmpfile);
    strcpy(tmpfile, settings.prj_name);
    strcat(tmpfile, ".9xz");
    delete(tmpfile);

    strcpy(tmpfile, settings.prj_name);
    strcat(tmpfile, ".89y");
    delete(tmpfile);
    strcpy(tmpfile, settings.prj_name);
    strcat(tmpfile, ".9xy");
    delete(tmpfile);

    return 0;
}

/*
//#define TEST
#ifdef TEST
    if (optind < argc)
    {
	printf ("non-option ARGV-elements: ");
	while (optind < argc)
	    printf ("%s ", argv[optind++]);
	printf ("\n");
    }
#else
    if (i < argc)
    {
	if(strstr(argv[i], ".tpr")) {
	    parse_file(argv[i]);
	}
	else {
	    usage(EXIT_FAILURE);
	    printf ("non-option ARGV-elements: ");
	    while (i < argc)
		printf ("%s ", argv[i++]);
	}
	printf ("\n");
    }
#endif
*/
